//
// $Id: output.cc,v 1.9 1999/11/29 15:11:18 voeckler Exp $
//
// Author:  Jens-S. Vckler <voeckler@rvs.uni-hannover.de>
//
// File:    output.cc
//          Thu Aug 05 1999
//
// (c) 1999 Lehrgebiet Rechnernetze und Verteilte Systeme
//          Universitt Hannover, Germany
//
// Permission to use, copy, modify, distribute, and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that (i) the above copyright notices and this permission
// notice appear in all copies of the software and related documentation,
// and (ii) the names of the Lehrgebiet Rechnernetze und Verteilte
// Systeme and the University of Hannover may not be used in any
// advertising or publicity relating to the software without the
// specific, prior written permission of Lehrgebiet Rechnernetze und
// Verteilte Systeme and the University of Hannover.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL THE LEHRGEBIET RECHNERNETZE UND VERTEILTE SYSTEME OR
// THE UNIVERSITY OF HANNOVER BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
// INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
// SOFTWARE.
//
// $Log: output.cc,v $
// Revision 1.9  1999/11/29 15:11:18  voeckler
// added new parameter to statistics output, in order to print statistics
// as comments into the database intermediate file.
//
// Revision 1.8  1999/10/29 14:08:38  voeckler
// made many functions module private, added entry point function
// printResults(), adjusted to moval of time stamps into counters.
//
// Revision 1.7  1999/09/02 10:20:48  voeckler
// added massive parallel look up facilities. The AS# lookup facilities
// are rather difficult.
//
// Revision 1.6  1999/08/27 21:51:27  voeckler
// changed String operator() to method c_str() call.
//
// Revision 1.5  1999/08/25 21:14:10  voeckler
// fixed yet another dns lookup bug, added distribution output.
//
// Revision 1.4  1999/08/22 11:52:56  voeckler
// added functionality to display duration as an alternative
// to data rates.
//
// Revision 1.3  1999/08/20 22:43:50  voeckler
// made dns cache a pointer. now prints full statistics of the
// irr and dns cache, if they are available.
//
// Revision 1.2  1999/08/15 22:31:44  voeckler
// The sublevel of the hierarchy overview are now sorted by requests.
//
// Revision 1.1  1999/08/05 21:17:11  voeckler
// Initial revision
//
//

#if defined(__GNUC__) || defined(__GNUG__)
#pragma implementation
#endif

#include <time.h>
#include <math.h> // modf()
#include <string.h>
#include <stdlib.h>
#include "ctype.hh"
#include "string.hh"
#include "stringset.hh"
#include "output.hh"
#include "tools.hh"
#include "global.hh"

static const char* RCS_ID =
"$Id: output.cc,v 1.9 1999/11/29 15:11:18 voeckler Exp $";

static String str_other( "<%d more>" );
static const char* timeFormat = "%d.%m.%Y %H:%M:%S";

typedef StringMap<String,false> StringStringMap;

inline
bool
isSymbolic( const String& s )
{ return ( DNSItem::aton(s) == C_U32(-1) ); }

inline
bool
isNumeric( const String& s )
{ return ( DNSItem::aton(s) != C_U32(-1) ); }

void
clientLookup( DNSCache* dns, const CountMap& clients, StringStringMap& result )
{
  // init result unless empty
  if ( result.size() ) result.clear();

  // combine all clients into a map for massive parallel lookups
  StringSet clientSet;
  for ( CountMap::Iterator i(clients); i.avail(); i++ ) {
    String key( i.key().lower() );
    if ( globals.clientFQDN && isNumeric(key) ) clientSet.insert(key);
    else result[key] = key;
  }

  // if there is something to be looked up, do it now
  if ( ! clientSet.empty() ) {
    DNSCache::DNSItemMap local( dns->getanybyany(clientSet) );
    for ( DNSCache::DNSItemMap::Iterator i(local); i.avail(); i++ ) {
      StringSet::iterator j = clientSet.find( i.key() );
      if ( j != clientSet.end() ) {
	DNSItem value( i.value() );
	result[i.key()] = value.empty ? i.key() : value.fqdn;
      }
    }
  }
}

// -------------------------------------------------------------------

static
void
showTitle( const char* cache, time_t start, time_t final )
{
  char s[80];
  strftime( s, sizeof(s), "%c %Z", localtime( &start ) );
  printf( "# Host \"%s\" from \"%s\" ", cache, s );
  strftime( s, sizeof(s), "%c %Z", localtime( &final ) );
  printf( "till \"%s\"\n\n", s );
}



static
void
showOverview( const Counters& counters )
{
  bool useRate( globals.preferDatarate );
  char s[64];
  Counter temp;

  puts( valueHead( "# SUM", 6, useRate, true ).c_str() );
  puts( valueTail( 6, true ).c_str() );
  printf( "UDP    %s\n", 
	  valueBody(counters.udp, counters.all, counters.udp_hit, useRate).c_str() );
  temp += counters.udp_hit;
  printf( "TCP    %s\n", 
	  valueBody(counters.tcp, counters.all, counters.tcp_hit, useRate).c_str() );
  temp += counters.tcp_hit;
  puts( valueTail( 6, true ).c_str() );
  printf( "ALL    %s\n", 
	  valueBody(counters.all, counters.all, temp, useRate).c_str() );
  puts("");
  
  puts( valueHead( "# UDP BY STATUS", 20, useRate, false ).c_str() );
  puts( valueTail( 20, false ).c_str() );
  printf( "%-20s %s\n", "HIT", 
	  valueBody(counters.udp_hit,counters.udp,useRate).c_str() );
  size_t size = LogLine::stable.distinct()+1; // include zeroth element!
  size_t* index = indexByReqs( counters.udp_hit_status, size );
  for ( MyUInt32 i=0; i<size; ++i )
    if ( counters.udp_hit_status[index[i]].reqs() )
      printf( " %-19s %s\n", 
	      LogLine::stable.reverse(s,sizeof(s),index[i]),
	      valueBody(counters.udp_hit_status[index[i]],
			counters.udp, useRate).c_str() );
  delete[] index;
  printf( "%-20s %s\n", "MISS", 
	  valueBody(counters.udp_miss,counters.udp,useRate).c_str() );
  index = indexByReqs( counters.udp_miss_status, size );
  for ( MyUInt32 i=0; i<size; ++i )
    if ( counters.udp_miss_status[index[i]].reqs() )
      printf( " %-19s %s\n", 
	      LogLine::stable.reverse(s,sizeof(s),index[i]),
	      valueBody(counters.udp_miss_status[index[i]],
			counters.udp, useRate).c_str() );
  delete[] index;
  puts( valueTail( 20, false ).c_str() );
  printf( "%-20s %s\n", "SUM", 
	  valueBody(counters.udp,counters.udp,useRate).c_str() );
  puts("");



  puts( valueHead( "# TCP BY STATUS", 20, useRate, false ).c_str() );
  puts( valueTail( 20, false ).c_str() );
  printf( "%-20s %s\n", "HIT", 
	  valueBody(counters.tcp_hit,counters.tcp,useRate).c_str() );
  index = indexByReqs( counters.tcp_hit_status, size );
  for ( MyUInt32 i=0; i<size; ++i )
    if ( counters.tcp_hit_status[index[i]].reqs() )
      printf( " %-19s %s\n", 
	      LogLine::stable.reverse(s,sizeof(s),index[i]),
	      valueBody(counters.tcp_hit_status[index[i]],
			counters.tcp,useRate).c_str() );
  delete[] index;
  printf( "%-20s %s\n", "MISS", 
	  valueBody(counters.tcp_miss,counters.tcp,useRate).c_str() );
  index = indexByReqs( counters.tcp_miss_status, size );
  for ( MyUInt32 i=0; i<size; ++i )
    if ( counters.tcp_miss_status[index[i]].reqs() )
      printf( " %-19s %s\n", 
	      LogLine::stable.reverse(s,sizeof(s),index[i]),
	      valueBody(counters.tcp_miss_status[index[i]],
			counters.tcp,useRate).c_str() );
  delete[] index;
  printf( "%-20s %s\n", "NONE", 
	  valueBody(counters.tcp_miss_none,counters.tcp,useRate).c_str() );
  index = indexByReqs( counters.tcp_miss_none_status, size );
  for ( MyUInt32 i=0; i<size; ++i )
    if ( counters.tcp_miss_none_status[index[i]].reqs() )
      printf( " %-19s %s\n", 
	      LogLine::stable.reverse(s,sizeof(s),index[i]),
	      valueBody(counters.tcp_miss_none_status[index[i]],
			counters.tcp,useRate).c_str() );
  delete[] index;
  puts( valueTail( 20, false ).c_str() );
  printf( "%-20s %s\n", "SUM", 
	  valueBody(counters.tcp,counters.tcp,useRate).c_str() );
  puts("");
}



extern "C" {
static
int 
myCompare( const void* a, const void* b )
{ 
  size_t aa = *((size_t*) a);
  size_t bb = *((size_t*) b);
  fprintf( stderr, "<< %d <=> %d >>\n", aa, bb );
  return aa - bb;
}
}

static
void
showHierarchy( DNSCache* dns, const Counters& counters )
{
  bool useRate( globals.preferDatarate );
  char s[80];
  Counter temp, thit;
  size_t size = LogLine::htable.distinct()+1; // include zeroth element!

  puts( valueHead( "# HIERARCHY OVERVIEW", 32, useRate, true ).c_str() );
  puts( valueTail( 32, true ).c_str() );
  printf( "%-32s %s\n", "DIRECT to source", 
	  valueBody( counters.hier_direct, counters.hier, useRate ).c_str() );
  size_t* index = indexByReqs( counters.hier_direct_method, size );
  for ( MyUInt32 i=0; i<size; ++i ) 
    if ( counters.hier_direct_method[index[i]].reqs() ) {
      printf( " %-31s %s\n", LogLine::htable.reverse(s,sizeof(s),index[i]),
	      valueBody( counters.hier_direct_method[index[i]],
			 counters.hier,
			 counters.tcp_hit_hier[index[i]],
			 useRate ).c_str() );
      temp += counters.hier_direct_method[index[i]];
      thit += counters.tcp_hit_hier[index[i]];
    }
  delete[] index;

  printf( "%-32s %s\n", "PARENT fetches", 
	  valueBody( counters.hier_parent, counters.hier, useRate ).c_str() );
  index = indexByReqs( counters.hier_parent_method, size );
  for ( MyUInt32 i=0; i<size; ++i ) 
    if ( counters.hier_parent_method[index[i]].reqs() ) {
      printf( " %-31s %s\n", LogLine::htable.reverse(s,sizeof(s),index[i]),
	      valueBody( counters.hier_parent_method[index[i]],
			 counters.hier,
			 counters.tcp_hit_hier[index[i]],
			 useRate ).c_str() );
      temp += counters.hier_parent_method[index[i]];
      thit += counters.tcp_hit_hier[index[i]];
    }
  delete[] index;

  printf( "%-32s %s\n", "PEER fetches", 
	  valueBody( counters.hier_peer, counters.hier, useRate ).c_str() );
  index = indexByReqs( counters.hier_peer_method, size );
  for ( MyUInt32 i=0; i<size; ++i ) 
    if ( counters.hier_peer_method[index[i]].reqs() ) {
      printf( " %-31s %s\n", LogLine::htable.reverse(s,sizeof(s),index[i]),
	      valueBody( counters.hier_peer_method[index[i]],
			 counters.hier,
			 counters.tcp_hit_hier[index[i]],
			 useRate ).c_str() );
      temp += counters.hier_peer_method[index[i]];
      thit += counters.tcp_hit_hier[index[i]];
    }
  delete[] index;

  puts( valueTail( 32, true ).c_str() );
  printf( "%-32s %s\n", "SUM", valueBody(temp,counters.hier,thit,useRate).c_str() );
  if ( temp != counters.hier )
    printf( "%-32s %s\n", "SHOULD-BE", 
	    valueBody(counters.hier,counters.hier,counters.tcp_hit,useRate).c_str() );
  puts("");

  // -----------------------------------------------------------------

  // combine all peers (neighbours + parents) into one map
  // for massive parallel DNS lookups of their respective address
  StringStringMap map;
  clientLookup( dns, counters.hier_peer_host, map );
  
  puts( valueHead( "# HIERARCHY BY HOST", 40, useRate, false ).c_str() );
  puts( valueTail( 40, false ).c_str() );
  printf( "%-40s %s\n", "DIRECT to source", 
	  valueBody( counters.hier_direct, counters.hier, useRate ).c_str() );

  index = indexByReqs( counters.hier_peer_host );
  Counter* sub = new Counter[size];
  temp = counters.hier_direct;
  for ( size_t i=0; i<counters.hier_peer_host.size(); ++i ) {
    String key( counters.hier_peer_host.key(index[i]) );
    printf( "%-40s %s\n", 
	    map[key.lower()].substring(0,40).c_str(),
	    valueBody( counters.hier_peer_host[index[i]], 
		       counters.hier, useRate ).c_str() );
    temp += counters.hier_peer_host[index[i]];
    
    for ( size_t j=0; j<size; ++j ) 
      sub[j] = counters.hier_peer_status[j][key];
    size_t* local = indexByReqs(sub,size);
    for ( size_t j=0; j<size; ++j ) {
      if ( sub[local[j]].reqs() ) 
	printf( " %-39s %s\n", LogLine::htable.reverse(s,sizeof(s),local[j]),
		valueBody( sub[local[j]], counters.hier, useRate ).c_str() );
    }
    delete[] local;
  }
  delete[] sub;
  delete[] index;
  puts( valueTail( 40, false ).c_str() );
  printf( "%-40s %s\n", "SUM", valueBody(temp,counters.hier,useRate).c_str() );
  if ( temp != counters.hier )
    printf( "%-40s %s\n", "SHOULD-BE", 
	    valueBody(counters.hier,counters.hier,useRate).c_str() );
  puts("");
}



static
void
showMethods( const Counters& counters )
{
  bool useRate(globals.preferDatarate);
  char s[64];
  Counter thit, temp;
  size_t size = LogLine::mtable.distinct()+1; // include zeroth element!
  size_t* index = indexByReqs( counters.methods, size );

  puts( valueHead("# METHODS", 10, useRate, true).c_str() );
  puts( valueTail(10,true).c_str() );
  for ( MyUInt32 i=0; i<size; ++i ) {
    if ( counters.methods[index[i]].reqs() ) {
      printf( "%-10s %s\n", LogLine::mtable.reverse(s,sizeof(s),index[i]), 
	      valueBody( counters.methods[index[i]],
			 counters.all,
			 counters.hit_methods[index[i]], useRate ).c_str() );
      temp += counters.methods[index[i]];
      thit += counters.hit_methods[index[i]];
    }
  }
  delete[] index;
  puts( valueTail(10,true).c_str() );
  printf( "%-10s %s\n\n", "SUM", valueBody(temp,counters.all,thit,useRate).c_str() );
  if ( temp != counters.all || thit != counters.tcp_hit+counters.udp_hit )
    printf( "%-10s %s\n\n", "SHOULD-BE", 
	    valueBody(counters.all,counters.all,
		      counters.tcp_hit+counters.udp_hit,useRate).c_str() );
}


static
void
showSchemes( const Counters& counters )
{
  bool useRate( globals.preferDatarate );
  char s[64];
  Counter thit, temp;
  size_t size = LogLine::schemes.distinct()+1; // include zeroth element!
  size_t* index = indexByReqs( counters.tcp_scheme, size );

  puts( valueHead("# SCHEMES", 16, useRate, true).c_str() );
  puts( valueTail(16,true).c_str() );
  for ( MyUInt32 i=0; i<size; ++i ) {
    if ( counters.tcp_scheme[index[i]].reqs() ) {
      printf( "%-16s %s\n", LogLine::schemes.reverse(s,sizeof(s),index[i]), 
	      valueBody( counters.tcp_scheme[index[i]],
			 counters.tcp,
			 counters.tcp_hit_scheme[index[i]],
			 useRate ).c_str() );
      temp += counters.tcp_scheme[index[i]];
      thit += counters.tcp_hit_scheme[index[i]];
    }
  }
  delete[] index;
  puts( valueTail(16,true).c_str() );
  printf( "%-16s %s\n\n", "SUM", valueBody(temp,counters.tcp,thit,useRate).c_str() );
  if ( temp != counters.tcp || thit != counters.tcp_hit )
    printf( "%-16s %s\n\n", "SHOULD-BE", 
	    valueBody(counters.tcp,counters.tcp,
		      counters.tcp_hit,useRate).c_str() );
}



static
void
showTLD( const Counters& counters, const size_t number )
{
  bool useRate( globals.preferDatarate );
  Counter thit;
  Counter temp;
  Counter leftover;
  Counter lhitover;
  MyUInt32 n = C_U32(0);

  puts( valueHead( "# TLD BY REQ.", 13, useRate, true ).c_str() );
  puts( valueTail( 13, true ).c_str() );
  size_t* index = indexByReqs( counters.tcp_tld );
  for ( size_t i=0; i<counters.tcp_tld.size(); ++i ) {
    String key( counters.tcp_tld.key(index[i]) );
    if ( number == 0 || number > i ) {
      printf( "%-13s %s\n", key.c_str(),
	      valueBody( counters.tcp_tld[index[i]], 
			 counters.tcp,
			 counters.tcp_hit_tld[key],
			 useRate ).c_str() );
    } else {
      leftover += counters.tcp_tld[index[i]];
      lhitover += counters.tcp_hit_tld[key];
      n++;
    }
    temp += counters.tcp_tld[index[i]];
    thit += counters.tcp_hit_tld[key];
  }
  delete[] index;

  if ( n ) {
    char here[16];
    sprintf( here, str_other.c_str(), n );
    printf( "%-13s %s\n", here, 
	    valueBody( leftover, counters.tcp, lhitover, useRate ).c_str() );
  }
  puts( valueTail( 13, true ).c_str() );
  printf( "%-13s %s\n", "SUM", valueBody(temp,counters.tcp,thit,useRate).c_str() );
  if ( temp != counters.tcp || thit != counters.tcp_hit )
    printf( "%-13s %s\n", "SHOULD-BE", 
	    valueBody(counters.tcp,counters.tcp,counters.tcp_hit,useRate).c_str() );
  puts("");




  puts( valueHead( "# TLD BY VOL.", 13, useRate, true ).c_str() );
  puts( valueTail( 13, true ).c_str() );
  index = indexBySize( counters.tcp_tld );
  thit = temp = Counter();
  leftover = lhitover = Counter();
  n = C_U32(0);
  for ( size_t i=0; i<counters.tcp_tld.size(); ++i ) {
    String key( counters.tcp_tld.key(index[i]) );
    if ( number == 0 || number > i ) {
      printf( "%-13s %s\n", key.c_str(),
	      valueBody( counters.tcp_tld[index[i]], 
			 counters.tcp,
			 counters.tcp_hit_tld[key],
			 useRate ).c_str() );
    } else {
      leftover += counters.tcp_tld[index[i]];
      lhitover += counters.tcp_hit_tld[key];
      n++;
    }
    temp += counters.tcp_tld[index[i]];
    thit += counters.tcp_hit_tld[key];
  }
  delete[] index;

  if ( n ) {
    char here[16];
    sprintf( here, str_other.c_str(), n );
    printf( "%-13s %s\n", here, 
	    valueBody( leftover, counters.tcp, lhitover, useRate ).c_str() );
  }
  puts( valueTail( 13, true ).c_str() );
  printf( "%-13s %s\n", "SUM", valueBody(temp,counters.tcp,thit,useRate).c_str() );
  if ( temp != counters.tcp || thit != counters.tcp_hit )
    printf( "%-13s %s\n", "SHOULD-BE", 
	    valueBody(counters.tcp,counters.tcp,counters.tcp_hit,useRate).c_str() );
  puts("");
}



static
void
showSLD( const Counters& counters, const size_t number )
{
  bool useRate( globals.preferDatarate );
  Counter thit;
  Counter temp;
  Counter leftover;
  Counter lhitover;
  MyUInt32 n = C_U32(0);

  puts( valueHead( "# 2ND LEVEL DOMAIN BY REQEST", 33, useRate, true ).c_str() );
  puts( valueTail( 33, true ).c_str() );
  size_t* index = indexByReqs( counters.tcp_sld );
  for ( size_t i=0; i<counters.tcp_sld.size(); ++i ) {
    String key( counters.tcp_sld.key(index[i]) );
    if ( number == 0 || number > i ) {
      printf( "%-33s %s\n", key.c_str(),
	      valueBody( counters.tcp_sld[index[i]], 
			 counters.tcp,
			 counters.tcp_hit_sld[key],
			 useRate ).c_str() );
    } else {
      leftover += counters.tcp_sld[index[i]];
      lhitover += counters.tcp_hit_sld[key];
      n++;
    }
    temp += counters.tcp_sld[index[i]];
    thit += counters.tcp_hit_sld[key];
  }
  delete[] index;

  if ( n ) {
    char here[16];
    sprintf( here, str_other.c_str(), n );
    printf( "%-33s %s\n", here, 
	    valueBody( leftover, counters.tcp, lhitover, useRate ).c_str() );
  }
  puts( valueTail( 33, true ).c_str() );
  printf( "%-33s %s\n", "SUM", valueBody(temp,counters.tcp,thit,useRate).c_str() );
  if ( temp != counters.tcp || thit != counters.tcp_hit )
    printf( "%-33s %s\n", "SHOULD-BE", 
	    valueBody(counters.tcp,counters.tcp,counters.tcp_hit,useRate).c_str() );
  puts("");




  puts( valueHead( "# 2ND LEVEL DOMAIN BY VOLUME", 33, useRate, true ).c_str() );
  puts( valueTail( 33, true ).c_str() );
  index = indexBySize( counters.tcp_sld );
  thit = temp = Counter();
  leftover = lhitover = Counter();
  n = C_U32(0);
  for ( size_t i=0; i<counters.tcp_sld.size(); ++i ) {
    String key( counters.tcp_sld.key(index[i]) );
    if ( number == 0 || number > i ) {
      printf( "%-33s %s\n", key.c_str(),
	      valueBody( counters.tcp_sld[index[i]], 
			 counters.tcp,
			 counters.tcp_hit_sld[key],
			 useRate ).c_str() );
    } else {
      leftover += counters.tcp_sld[index[i]];
      lhitover += counters.tcp_hit_sld[key];
      n++;
    }
    temp += counters.tcp_sld[index[i]];
    thit += counters.tcp_hit_sld[key];
  }
  delete[] index;

  if ( n ) {
    char here[16];
    sprintf( here, str_other.c_str(), n );
    printf( "%-33s %s\n", here, 
	    valueBody( leftover, counters.tcp, lhitover, useRate ).c_str() );
  }
  puts( valueTail( 33, true ).c_str() );
  printf( "%-33s %s\n", "SUM", valueBody(temp,counters.tcp,thit,useRate).c_str() );
  if ( temp != counters.tcp || thit != counters.tcp_hit )
    printf( "%-33s %s\n", "SHOULD-BE", 
	    valueBody(counters.tcp,counters.tcp,counters.tcp_hit,useRate).c_str() );
  puts("");
}



static
void
showMediatypes( const Counters& counters )
{
  bool useRate( globals.preferDatarate );
  char s[80];
  Counter thit;
  Counter temp;

  puts( valueHead( "# MEDIATYPES BY REQUESTS", 24, useRate, true ).c_str() );
  puts( valueTail( 24, true ).c_str() );
  size_t* index = indexByReqs( counters.tcp_mime, counters.nMedia );
  for ( size_t i=0; i<counters.nMedia; ++i ) {
    if ( counters.tcp_mime[index[i]].reqs() ) {
      printf( "%-24s %s\n", 
	      LogLine::mediatypes.reverse(s,sizeof(s),index[i]),
	      valueBody( counters.tcp_mime[index[i]], 
			 counters.tcp,
			 counters.tcp_hit_mime[index[i]],
			 useRate ).c_str() );
      if ( counters.tcp_submime[index[i]] ) {
	size_t* sub = indexByReqs( counters.tcp_submime[index[i]],
				   counters.nSubtype[index[i]] );
	for ( size_t j=0; j<counters.nSubtype[index[i]]; ++j ) {
	  if ( counters.tcp_submime[index[i]][sub[j]].reqs() ) {
	    printf( " /%-22s %s\n",
		    LogLine::subtypes[index[i]].reverse(s,sizeof(s),sub[j]),
		    valueBody( counters.tcp_submime[index[i]][sub[j]],
			       counters.tcp,
			       counters.tcp_hit_submime[index[i]][sub[j]],
			       useRate ).c_str() );
	  }
	}
	delete[] sub;
      }
    }
    temp += counters.tcp_mime[index[i]];
    thit += counters.tcp_hit_mime[index[i]];
  }
  delete[] index;
  puts( valueTail( 24, true ).c_str() );
  printf( "%-24s %s\n", "SUM", valueBody(temp,counters.tcp,thit,useRate).c_str() );
  if ( temp != counters.tcp || thit != counters.tcp_hit )
    printf( "%-24s %s\n", "SHOULD-BE", 
	    valueBody(counters.tcp,counters.tcp,counters.tcp_hit,useRate).c_str() );
  puts("");




  puts( valueHead( "# MEDIATYPES BY VOLUME", 24, useRate, true ).c_str() );
  puts( valueTail( 24, true ).c_str() );
  index = indexBySize( counters.tcp_mime, counters.nMedia );
  thit = temp = Counter();
  for ( size_t i=0; i<counters.nMedia; ++i ) {
    if ( counters.tcp_mime[index[i]].reqs() ) {
      printf( "%-24s %s\n", 
	      LogLine::mediatypes.reverse(s,sizeof(s),index[i]),
	      valueBody( counters.tcp_mime[index[i]], 
			 counters.tcp,
			 counters.tcp_hit_mime[index[i]],
			 useRate ).c_str() );
      if ( counters.tcp_submime[index[i]] ) {
	size_t* sub = indexBySize( counters.tcp_submime[index[i]],
				   counters.nSubtype[index[i]] );
	for ( size_t j=0; j<counters.nSubtype[index[i]]; ++j ) {
	  if ( counters.tcp_submime[index[i]][sub[j]].reqs() ) {
	    printf( " /%-22s %s\n",
		    LogLine::subtypes[index[i]].reverse(s,sizeof(s),sub[j]),
		    valueBody( counters.tcp_submime[index[i]][sub[j]],
			       counters.tcp,
			       counters.tcp_hit_submime[index[i]][sub[j]],
			       useRate ).c_str() );
	  }
	}
	delete[] sub;
      }
    }
    temp += counters.tcp_mime[index[i]];
    thit += counters.tcp_hit_mime[index[i]];
  }
  delete[] index;
  puts( valueTail( 24, true ).c_str() );
  printf( "%-24s %s\n", "SUM", valueBody(temp,counters.tcp,thit,useRate).c_str() );
  if ( temp != counters.tcp || thit != counters.tcp_hit )
    printf( "%-24s %s\n", "SHOULD-BE", 
	    valueBody(counters.tcp,counters.tcp,counters.tcp_hit,useRate).c_str() );
  puts("");
}



static
void
showPorts( const Counters& counters )
{
  bool useRate( globals.preferDatarate );
  char temp[80];
  Counter sum;
  // with without port 80, which is stored in ports[64]...
  for ( int i=0; i<64; ++i ) sum += counters.tcp_ports[i];
  puts( valueHead( "# NON-STD PORTS", 15, useRate, false ).c_str() );
  puts( valueTail( 15, false ).c_str() );
  for ( int i=0; i<64; ++i )
    if ( counters.tcp_ports[i].reqs() ) {
      sprintf( temp, "%5d-%-5d", i << 10, ((i+1)<<10)-1 );
      if ( i==0 ) strcat( temp, "\\80" );
      printf( "%-15s %s\n", temp, 
	      valueBody( counters.tcp_ports[i], sum, useRate ).c_str() );
    }
  puts( valueTail( 15, false ).c_str() );
  Counter total = sum + counters.tcp_ports[64];
  printf( "%-15s %s\n", "NON-STD SUM", 
	  valueBody( sum, total, useRate ).c_str() );
  printf( "%-15s %s\n", "Port 80", 
	  valueBody( counters.tcp_ports[64], total, useRate ).c_str() );
  puts( valueTail( 15, false ).c_str() );
  printf( "%-15s %s\n", "SUM", valueBody( total, total, useRate ).c_str() );
  puts("");
}


static
void
showUDPClients( DNSCache* dns, const Counters& counters, const size_t number )
{
  bool useRate( globals.preferDatarate );
  Counter thit;
  Counter temp;
  Counter leftover;
  Counter lhitover;
  MyUInt32 n = C_U32(0);

  // combine all UDP clients into a map for massive parallel lookups
  StringStringMap map;
  clientLookup( dns, counters.udp_client, map );
  
  puts( valueHead( "# UDP CLIENTS BY REQUESTS", 40, useRate, true ).c_str() );
  puts( valueTail( 40, true ).c_str() );
  size_t size = counters.udp_client.size();
  size_t* index = indexByReqs( counters.udp_client );
  for ( size_t i=0; i<size; ++i ) {
    if ( counters.udp_client[index[i]].reqs() ) {
      String key( counters.udp_client.key(index[i]) );
      if ( number == 0 || number > i ) {
	printf( "%-40s %s\n", map[key.lower()].substring(0,40).c_str(),
		valueBody( counters.udp_client[index[i]], 
			   counters.udp,
			   counters.udp_hit_client[key],
			   useRate ).c_str() );
      } else {
	leftover += counters.udp_client[index[i]];
	lhitover += counters.udp_hit_client[key];
	n++;
      }
      temp += counters.udp_client[index[i]];
      thit += counters.udp_hit_client[key];
    }
  }
  delete[] index;

  if ( n ) {
    char here[80];
    sprintf( here, str_other.c_str(), n );
    printf( "%-40s %s\n", here,
	    valueBody( leftover, counters.udp, lhitover, useRate ).c_str() );
  }
  puts( valueTail( 40, true ).c_str() );
  printf( "%-40s %s\n", "SUM", valueBody(temp,counters.udp,thit,useRate).c_str() );
  if ( temp != counters.udp || thit != counters.udp_hit )
    printf( "%-40s %s\n", "SHOULD-BE", 
	    valueBody(counters.udp,counters.udp,counters.udp_hit,useRate).c_str() );
  puts("");
}


static
void
showTCPClients( DNSCache* dns, const Counters& counters, const size_t number )
{
  bool useRate( globals.preferDatarate );
  Counter  temp,  leftover;
  Counter  thit,  lhitover;
  Counter  terr,  lerrover;
  Counter tmiss, lmissover;
  MyUInt32 n = C_U32(0);

  // combine all TCP clients into a map for massive parallel lookups
  StringStringMap map;
  clientLookup( dns, counters.tcp_client, map );

  puts( valueHead( "# TCP CLIENTS BY REQUST", 40, useRate, true ).substitute("HIT","HME",true).c_str() );
  puts( valueTail( 40, true ).c_str() );

  size_t size = counters.tcp_client.size();
  size_t* index = indexByReqs( counters.tcp_client );
  for ( size_t i=0; i<size; ++i ) {
    if ( counters.tcp_client[index[i]].reqs() ) {
      String key( counters.tcp_client.key(index[i]) );
      if ( number == 0 || number > i ) {
	printf( "%-40s %s\n", map[key.lower()].substring(0,40).c_str(),
		valueBody( counters.tcp_client[index[i]], 
			   counters.tcp,
			   counters.tcp_hit_client[key],
			   useRate ).c_str() );
	if ( counters.tcp_miss_client[key].reqs() )
	  printf( " %-83s %s\n", "MISS",
		  valueBody( counters.tcp_miss_client[key],
			     counters.tcp_client[index[i]],
			     useRate ).c_str() );
	if ( counters.tcp_miss_none_client[key].reqs() )
	  printf( " %-83s %s\n", "NONE",
		  valueBody( counters.tcp_miss_none_client[key],
			     counters.tcp_client[index[i]],
			     useRate ).c_str() );
      } else {
	leftover += counters.tcp_client[index[i]];
	lhitover += counters.tcp_hit_client[key];
	lmissover += counters.tcp_miss_client[key];
	lerrover += counters.tcp_miss_none_client[key];
	n++;
      }
      temp += counters.tcp_client[index[i]];
      thit += counters.tcp_hit_client[key];
      tmiss += counters.tcp_miss_client[key];
      terr += counters.tcp_miss_none_client[key];
    }
  }
  delete[] index;

  if ( n ) {
    char here[80];
    sprintf( here, str_other.c_str(), n );
    printf( "%-40s %s\n", here, 
	    valueBody(leftover,counters.tcp,lhitover,useRate).c_str() );
    if ( lmissover.reqs() )
      printf( " %-83s %s\n", "MISS", 
	      valueBody(lmissover,counters.tcp,useRate).c_str() );
    if ( lerrover.reqs() )
      printf( " %-83s %s\n", "NONE", 
	      valueBody(lerrover,counters.tcp,useRate).c_str() );
  }

  puts( valueTail( 40, true ).c_str() );
  printf( "%-40s %s\n", "SUM", 
	  valueBody(temp,counters.tcp,thit,useRate).c_str() );
  if ( tmiss.reqs() )
    printf( " %-83s %s\n", "MISS-SUM", 
	    valueBody(tmiss,counters.tcp,useRate).c_str() );
  if ( terr.reqs() )
    printf( " %-83s %s\n", "NONE-SUM", 
	    valueBody(terr,counters.tcp,useRate).c_str() );
  if ( temp != counters.tcp || thit != counters.tcp_hit ||
       terr != counters.tcp_miss_none || tmiss != counters.tcp_miss ) {
    printf( "%-40s %s\n", "SHOULD-BE", 
	    valueBody(counters.tcp,counters.tcp,counters.tcp_hit,useRate).c_str() );
    if ( counters.tcp_miss.reqs() )
      printf( " %-83s %s\n", "SHOULD-MISS", 
	      valueBody(counters.tcp_miss,counters.tcp,useRate).c_str() );
    if ( counters.tcp_miss_none.reqs() )
      printf( " %-83s %s\n", "SHOULD-NONE", 
	      valueBody(counters.tcp_miss_none,counters.tcp,useRate).c_str() );
  }
  puts("");
}

const String no_dns( "<NODNS>" );
const String no_irr( "<NO_AS>" );

void
hierarchyHostLookup( DNSCache* dns,
		     const CountMap& clients, AddressMap& result )
{
  // init result unless empty
  if ( result.size() ) result.clear();

  // combine all addresses into a map for massive parallel lookups
  StringSet clientSet;
  for ( CountMap::Iterator i(clients); i.avail(); i++ ) {
    String key( i.key().lower() );
    MyUInt32 address = DNSItem::aton(key);
    if ( address == C_U32(-1) ) clientSet.insert(key);
    else result[key] = address;
  }

  // if there is something to be looked up, do it now
  if ( ! clientSet.empty() ) {
    DNSCache::DNSItemMap local( dns->getanybyany(clientSet) );
    for ( DNSCache::DNSItemMap::Iterator i(local); i.avail(); i++ ) {
      StringSet::iterator j = clientSet.find( i.key() );
      if ( j != clientSet.end() ) {
	DNSItem value( i.value() );
	result[i.key()] = ( value.empty || value.n_addresses == 0 ) ?
	  C_U32(-1) :
	  value.addresses[0];
      }
    }
  }
}

void
hierarchyASLookup( IRRCache* irr, 
		   const CountMap& clients, const AddressMap& addr_map, 
		   IRRCache::IRRItemMap& dest_map, CountMap& asnmap )
{
  // create ASN map
  StringSet destination;
  for ( CountMap::Iterator i(clients); i.avail(); i++ ) {
    if ( i.value().reqs() ) {
      MyUInt32 address = addr_map[ i.key().lower() ];
      if ( address != C_U32(-1) ) {
	// DNS success, add host address to lookup for IRR
	destination.insert(DNSItem::ntoa(address));
      }
    }
  }

  // massive parallel IRR lookup, address to origin
  dest_map = irr->getOriginByAddress( destination );
  // The answers in dest_map contain the host address,
  // NOT the network/CIDR address.

  // count all items, while collecting AS numbers in "asnumber" for 
  // further lookups of the AS description
  StringSet asnumber;
  for ( CountMap::Iterator i(clients); i.avail(); i++ ) {
    if ( i.value().reqs() ) {
      MyUInt32 address = addr_map[ i.key().lower() ];
      if ( address != C_U32(-1) ) {
	// DNS success, now check IRR
	IRRItem what = dest_map[ DNSItem::ntoa(address) ];
	if ( what.empty ) {
	  // IRR failure
	  asnmap[ no_irr ] += i.value();
	} else {
	  // IRR success
	  asnmap[what.item] += i.value();
	  asnumber.insert(what.item);
	}
      } else {
	// DNS failure
	asnmap[ no_dns ] += i.value();
      }
    }
  }

  // massive parallel IRR lookup, origin to description
  dest_map += irr->getanybyany( asnumber );
}

static
void
showASN( DNSCache* dns, IRRCache* irr, 
	 const Counters& counters, const size_t number )
{
  bool useRate( globals.preferDatarate );
  Counter temp;
  Counter leftover;
  MyUInt32 n = C_U32(0);
  CountMap asnmap;
  String host;

  // combine all HIER hosts into a map for massive parallel DNS lookups
  AddressMap addr_map;
  hierarchyHostLookup( dns, counters.hier_direct_host, addr_map );
  // POSTCONDITION: addr_map contains all addresses belonging to the 
  // direct hosts. If the HNo patch was applied, the address conversion
  // will be done without needing to bother DNS (much faster).

  // combine all HIER addresses from HIER hosts for massive parallel IRR
  IRRCache::IRRItemMap dest_map;
  hierarchyASLookup( irr, counters.hier_direct_host, addr_map,
		     dest_map, asnmap );
  // POSTCONDITION: dest_map will contain *hosts* key (not network/CIDR)
  // as dotted quad derived from addr_map which got its hosts from the
  // hier_direct_host map. The value type is the AS#, <NO_AS>, or <NODNS>.
  // It will also contain for the key AS# the description of the AS.
  // asnmap is a count map

  puts( valueHead( "# DIRECT DST BY REQUESTS", 40, useRate, false ).c_str() );
  puts( valueTail( 40, false ).c_str() );
  size_t size = asnmap.size();
  size_t* index = indexByReqs( asnmap );
  for ( size_t i=0; i<size; ++i ) {
    String key(asnmap.key(index[i]));
    if ( number == 0 || number > i ) {
      printf( "%-7s \"%-30s\" %s\n", key.c_str(), 
	      dest_map[key].item.substring(0,30).c_str(),
	      valueBody( asnmap[size_t(index[i])], counters.hier, useRate ).c_str() );
    } else {
      leftover += asnmap[size_t(index[i])];
      n++;
    }
    temp += asnmap[index[i]];
  }
  delete[] index;

  if ( n ) {
    char here[80];
    sprintf( here, str_other.c_str(), n );
    printf( "%-40s %s\n", here, 
	    valueBody( leftover, counters.hier_direct, useRate ).c_str() );
  }
  puts( valueTail( 40, false ).c_str() );
  printf( "%-40s %s\n", "SUM", 
	  valueBody(temp,counters.hier_direct,useRate).c_str() );
  if ( temp != counters.hier_direct )
    printf( "%-40s %s\n", "SHOULD-BE", 
	    valueBody(counters.hier_direct,
		      counters.hier_direct,
		      useRate).c_str() );
  puts("");



  temp = leftover = Counter();
  n = C_U32(0);
  puts( valueHead( "# DIRECT DST BY VOLUME", 40, useRate, false ).c_str() );
  puts( valueTail( 40, false ).c_str() );
  index = indexBySize( asnmap );
  for ( size_t i=0; i<size; ++i ) {
    String key(asnmap.key(index[i]));
    if ( number == 0 || number > i ) {
      printf( "%-7s \"%-30s\" %s\n", key.c_str(),
	      dest_map[key].item.substring(0,30).c_str(),
	      valueBody( asnmap[size_t(index[i])], 
			 counters.hier, 
			 useRate ).c_str() );
    } else {
      leftover += asnmap[size_t(index[i])];
      n++;
    }
    temp += asnmap[index[i]];
  }
  delete[] index;

  if ( n ) {
    char here[80];
    sprintf( here, str_other.c_str(), n );
    printf( "%-40s %s\n", here, 
	    valueBody( leftover, counters.hier_direct, useRate ).c_str() );
  }
  puts( valueTail( 40, false ).c_str() );
  printf( "%-40s %s\n", "SUM", 
	  valueBody(temp,counters.hier_direct,useRate).c_str() );
  if ( temp != counters.hier_direct )
    printf( "%-40s %s\n", "SHOULD-BE", 
	    valueBody(counters.hier_direct,
		      counters.hier_direct,
		      useRate).c_str() );
  puts("");
}

static
void
showPeaks( const Counters& counters )
{
  char tbuf[256];

  printf( "%-19s %8s %8s      %8s %8s      %8s      %8s      %8s\n",
	  "# REQUEST PEAKS", "udp", "udp-hit", "tcp", "tcp-hit",
	  "direct", "parent", "peer" );
  puts( "------------------- -------- ------------- -------- ------------- ------------- ------------- -------------" );
  for ( MyUInt32 i=counters.startTime / globals.peakInterval;
	i <= counters.finalTime / globals.peakInterval; 
	++i ) {
    time_t then = i * globals.peakInterval;
    strftime( tbuf, sizeof(tbuf), timeFormat, localtime(&then) );
    printf( "%s " SF_wU32(8) " " SF_wU32(8) "%s" SF_wU32(8) " "
	    SF_wU32(8) "%s" SF_wU32(8) "%s" SF_wU32(8) "%s" SF_wU32(8) "%s\n",
	    tbuf,
	    counters.peak_udp[i].reqs(), 
	    counters.peak_udp_hit[i].reqs(),
	    autoPercent( counters.peak_udp[i].reqs(),
			 counters.peak_udp_hit[i].reqs() ).c_str(),
	    counters.peak_tcp[i].reqs(), 
	    counters.peak_tcp_hit[i].reqs(),
	    autoPercent( counters.peak_tcp[i].reqs(),
			 counters.peak_tcp_hit[i].reqs() ).c_str(),
	    counters.peak_hier_direct[i].reqs(),
	    autoPercent( counters.peak_tcp[i].reqs(), 
			 counters.peak_hier_direct[i].reqs() ).c_str(),
	    counters.peak_hier_parent[i].reqs(),
	    autoPercent( counters.peak_tcp[i].reqs(), 
			 counters.peak_hier_parent[i].reqs() ).c_str(),
	    counters.peak_hier_peer[i].reqs(),
	    autoPercent( counters.peak_tcp[i].reqs(), 
			 counters.peak_hier_peer[i].reqs() ).c_str() );
  }
  puts("");

  printf( "%-19s %10s %10s      %10s %10s      %10s      %10s      %10s\n",
	  "# VOLUME PEAKS", "udp", "udp-hit", "tcp", "tcp-hit",
	  "direct", "parent", "peer" );
  puts( "------------------- ---------- --------------- ---------- --------------- --------------- --------------- ---------------" );
  for ( MyUInt32 i=counters.startTime / globals.peakInterval;
	i <= counters.finalTime / globals.peakInterval; 
	++i ) {
    time_t then = i * globals.peakInterval;
    strftime( tbuf, sizeof(tbuf), timeFormat, localtime(&then) );
    printf( "%s %s %s%s%s %s%s%s%s%s%s%s%s\n",
	    tbuf,
	    autoSize( counters.peak_udp[i].size() ).c_str(), 
	    autoSize( counters.peak_udp_hit[i].size() ).c_str(),
	    autoPercent( counters.peak_udp[i].size(),
			 counters.peak_udp_hit[i].size() ).c_str(),
	    autoSize( counters.peak_tcp[i].size() ).c_str(), 
	    autoSize( counters.peak_tcp_hit[i].size() ).c_str(),
	    autoPercent( counters.peak_tcp[i].size(),
			 counters.peak_tcp_hit[i].size() ).c_str(),
	    autoSize( counters.peak_hier_direct[i].size() ).c_str(),
	    autoPercent( counters.peak_tcp[i].size(), 
			 counters.peak_hier_direct[i].size() ).c_str(),
	    autoSize( counters.peak_hier_parent[i].size() ).c_str(),
	    autoPercent( counters.peak_tcp[i].size(), 
			 counters.peak_hier_parent[i].size() ).c_str(),
	    autoSize( counters.peak_hier_peer[i].size() ).c_str(),
	    autoPercent( counters.peak_tcp[i].size(), 
			 counters.peak_hier_peer[i].size() ).c_str() );
  }
  puts("");

  printf( "%-19s %10s%10s      %10s      %10s      %10s\n",
	  "# DURATION PEAKS", "tcp", "tcp-hit", "direct", "parent", "peer" );
  puts( "------------------- ---------- --------------- --------------- --------------- ---------------" );
  for ( MyUInt32 i=counters.startTime / globals.peakInterval;
	i <= counters.finalTime / globals.peakInterval; 
	++i ) {
    time_t then = i * globals.peakInterval;
    strftime( tbuf, sizeof(tbuf), timeFormat, localtime(&then) );
    printf( "%s %s%s%s%s%s%s%s%s%s\n",
	    tbuf,
	    autoTime( counters.peak_tcp[i].time(), 7, 1 ).c_str(), 
	    autoTime( counters.peak_tcp_hit[i].time(), 7, 1 ).c_str(),
	    autoPercent( counters.peak_tcp[i].time(),
			 counters.peak_tcp_hit[i].time() ).c_str(),
	    autoTime( counters.peak_hier_direct[i].time(), 7, 1 ).c_str(),
	    autoPercent( counters.peak_tcp[i].time(), 
			 counters.peak_hier_direct[i].time() ).c_str(),
	    autoTime( counters.peak_hier_parent[i].time(), 7, 1 ).c_str(),
	    autoPercent( counters.peak_tcp[i].time(), 
			 counters.peak_hier_parent[i].time() ).c_str(),
	    autoTime( counters.peak_hier_peer[i].time(), 7, 1 ).c_str(),
	    autoPercent( counters.peak_tcp[i].time(), 
			 counters.peak_hier_peer[i].time() ).c_str() );
  }
  puts("");
}

static
void
showDistribution( const Counters& counters )
{
  static const char* msg[Counters::D_END] = { "HIT", "MISS", "NONE" };

  for ( int d=Counters::D_HIT; d<Counters::D_END; ++d ) {
    printf( "# TCP %s DISTRIBUTION\n", msg[d] );

    // find start and stop of columns
    int sizes[2] = { 0, 31 };
    while ( counters.distribution[d].sizeDist[sizes[0]] == 0 && 
	    sizes[0] <= sizes[1] ) sizes[0]++;
    while ( counters.distribution[d].sizeDist[sizes[1]] == 0 &&
	    sizes[0] <= sizes[1] ) sizes[1]--;

    // find start and stop of lines
    int times[2] = { 0, 31 };
    while ( counters.distribution[d].timeDist[times[0]] == 0 && 
	    times[0] <= times[1] ) times[0]++;
    while ( counters.distribution[d].timeDist[times[1]] == 0 &&
	    times[0] <= times[1] ) times[1]--;
    printf( "%10s:", "duration" );

    // sanity check
    if ( times[1] < times[0] || sizes[1] < sizes[0] ) {
      puts( "<empty>" );
      continue;
    }

    // print header
    for ( int s=sizes[0]; s<=sizes[1]; ++s ) {
      String msg( autoSize( s ? pow(2.0,s-1) : 0, 999.9, 5, 0 ) );
      printf( "%6s", msg.substitute(" ","",true).c_str() );
    }
    puts( " = TIME-SUM" );

    // print tail
    fputs( " ----------", stdout );
    for ( int s=sizes[0]; s<=sizes[1]; ++s ) fputs( " -----", stdout );
    puts( "   --------" );

    for ( int t=times[0]; t<=times[1]; ++t ) {
      printf( " %s:", autoTime( t ? pow(2.0,t-1) : 0, 6, 1 ).c_str() );
      for ( int s=sizes[0]; s<=sizes[1]; ++s ) {
	printf( " %5lu", counters.distribution[d].timeSizeDist[t][s] );
      }
      printf( " = %8lu\n", counters.distribution[d].timeDist[t] );
    }

    // print tail
    fputs( " ----------", stdout );
    for ( int s=sizes[0]; s<=sizes[1]; ++s ) fputs( " -----", stdout );
    puts( "   --------" );

    // print size sums
    MyUInt64 sum = C_U64(0);
    printf( "%10s:", "SIZE-SUM" );
    for ( int s=sizes[0]; s<=sizes[1]; ++s ) {
      printf( " %5lu", counters.distribution[d].sizeDist[s] );
      sum += counters.distribution[d].sizeDist[s];
    }
    printf( " = " SF_wU64(8) "\n\n", sum );
  }
}

static
String
sftime( double stamp )
{
  char temp[3][32];
  double integer;
  sprintf( temp[0], "%.3f", modf( stamp, &integer ) );
  time_t what = (time_t) integer;
  
  strftime( temp[1], 32, "%H:%M:%S", localtime(&what) );
  strcpy( temp[2], temp[1] );
  strcat( temp[2], temp[0]+1 );
  return String(temp[2]);
}

static
void
cacheStats( FILE* out, const char* prefix,
	    const char* name, size_t size, const BaseCache* cache )
{
  MyUInt32 all = cache->request(BaseCache::C_ALL);
  MyUInt32 hit = cache->request(BaseCache::C_HIT);
  fprintf( out, "%s%s cache (%u items) had " SF_U32 " requests with " 
	   SF_U32 " hits (%s),\n", prefix,
	   name, size, all, hit, autoPercent(all,hit).trim().c_str() );

  MyUInt32 miss = cache->request(BaseCache::C_MISS);
  MyUInt32 ttl = cache->request(BaseCache::C_TTL);
  MyUInt32 late = cache->request(BaseCache::C_LATE);
  fprintf( out, "%s " SF_U32 " ttl expiries (%s) + "
	   SF_U32 " cache misses (%s) - "
	   SF_U32 " late hits (%s),\n", prefix,
	   ttl, autoPercent(all,ttl).trim().c_str(), 
	   miss, autoPercent(all,miss).trim().c_str(),
	   late, autoPercent(all,late).trim().c_str() );

  MyUInt32 pos = cache->request(BaseCache::C_POS);
  MyUInt32 neg = cache->request(BaseCache::C_NEG);
  fprintf( out, "%s resulting in " SF_U32 " successful (%s) and " SF_U32 
	   " failed (%s) lookups.\n", prefix,
	   pos, autoPercent(all,pos).trim().c_str(),
	   neg, autoPercent(all,neg).trim().c_str() );
}

void
showStats( FILE* out, const char* prefix,
	   const char* cache, DNSCache* dns, IRRCache* irr,
	   double looptime, double startup, MyUInt32 lineno )
{
  double runtime = now();
  fprintf( out, "%s# STATISTICS FOR \"%s\"\n", prefix, cache );
  fprintf( out, "%sanalysis started at %s\n", prefix, 
	   sftime(startup).c_str() );
  fprintf( out, "%sparsing finished at %s\n", prefix, 
	   sftime(startup+looptime).c_str() );
  fprintf( out, "%sresults printed at %s\n", prefix, 
	   sftime(runtime).c_str() );

  runtime -= startup;
  fprintf( out, "%s%.1f s spent in analyser\n", prefix, runtime );
  fprintf( out, "%s%.1f s spent in parser loop\n",
	   prefix, looptime );
  fprintf( out, "%s" SF_U32 " lines processed (%.1f lps)\n", 
	   prefix, lineno, lineno / looptime );
  
  if ( dns ) cacheStats( out, prefix, "DNS", dns->size(), dns );
  if ( irr ) cacheStats( out, prefix, "IRR", irr->size(), irr );
  fputc( '\n', out );
}

// ----------------------------------------------------------------------

void
printResults( const char* cache,
	      DNSCache* dns, IRRCache* irr, const Counters& counters,
	      double looptime, double startup, MyUInt32 lineno )
  // purpose: print results
  // paramtr: cache (IN): (virtual) name to use for output
  //          dns (IO): DNS query and cache
  //          irr (IO): IRR query and cache
  //          counters (IN): current value of last log file processing
  //          looptime (IN): time spent parsing the log file
  //          startup (IN): start time of seafood
  //          lineno (IN): number of lines processed.
{
  fputs( "# title...\n", stderr );
  showTitle( cache, counters.startTime, counters.finalTime );
  fflush(stdout);

  // overview
  fputs( "# overview...\n", stderr );
  showOverview(counters);
  fflush(stdout);

  // hierarchy info
  fputs( "# hierarchy...\n", stderr );
  showHierarchy(dns,counters);
  fflush(stdout);

  // methods
  fputs( "# methods...\n", stderr );
  showMethods(counters);
  fflush(stdout);

  // schemes
  fputs( "# schemes...\n", stderr );
  showSchemes(counters);
  fflush(stdout);

  // top-level domains
  fputs( "# tlds...\n", stderr );
  showTLD( counters, globals.tableSize[Global::S_TLD] );
  fflush(stdout);

  // 2nd-level domains
  fputs( "# 2lds...\n", stderr );
  showSLD( counters, globals.tableSize[Global::S_2LD] );
  fflush(stdout);

  if ( dns && irr ) {
    // show directs
    fputs( "# asn...\n", stderr );
    showASN( dns, irr, counters, globals.tableSize[Global::S_ASN] );
    fflush(stdout);
  }

  // media types
  fputs( "# mime...\n", stderr );
  showMediatypes(counters);
  fflush(stdout);

  // show port statistics
  fputs( "# ports...\n", stderr );
  showPorts( counters );
  fflush(stdout);

  // show clients
  fputs( "# clients...\n", stderr );
  showUDPClients( dns, counters, globals.tableSize[Global::S_UDP] );
  fflush(stdout);
  showTCPClients( dns, counters, globals.tableSize[Global::S_TCP] );
  fflush(stdout);

  // show peak intervals
  fputs( "# peaks...\n", stderr );
  showPeaks( counters );
  fflush(stdout);

  // show distributions
  if ( globals.showDistributions ) {
    fputs( "# distributions\n", stderr );
    showDistribution( counters );
    fflush(stdout);
  }

  // statistics about the log file analysis
  fputs( "# stats...\n", stderr );
  if ( looptime < 1E-4 ) looptime = 1E-4;
  showStats( stdout, "", cache, dns, irr, looptime, startup, lineno );
  fflush(stdout);
}
